//////////////////////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief Provide an interface matching BASISFIELD::splinefield, specifically for use with
///        EDDY::MoveBySuscCF class
/// \details Essentially this is a wrapper around a BASISFIELD::splinefield object, which
///          re-implements some of the more expensive calculations on the GPU, whilst simply
///          passing other calculations through to the underlying BASISFIELD::splinefield
///          object. As such, expect limited functionality as compared to the BASISFIELD
///          version - just enough to get EDDY::MoveBySuscCF running faster. Additionally,
///          we are using the pimpl idiom here.
/// \author Frederik Lange
/// \date February 2018
/// \copyright Copyright (C) 2018 University of Oxford
//////////////////////////////////////////////////////////////////////////////////////////////

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#ifndef CBF_SPLINE_FIELD_CUH
#define CBF_SPLINE_FIELD_CUH

#include <vector>
#include <memory>

#include "newimage/newimageall.h"
#include "armawrap/newmat.h"
#include "miscmaths/bfmatrix.h"
#include "basisfield/basisfield.h"
#include "basisfield/splinefield.h"

/// Multi-Modal Registration Framework
namespace CBF
{
  class CBFSplineField
  {
    public:
      /// Default dtor
      ~CBFSplineField();
      /// Move ctor
      CBFSplineField(CBFSplineField&& rhs);
      /// Move assignment operator
      CBFSplineField& operator=(CBFSplineField&& rhs);
      /// Copy ctor
      CBFSplineField(const CBFSplineField& rhs);
      /// Copy assignment operator
      CBFSplineField& operator=(const CBFSplineField& rhs);

      // Methods to wrap
      /// Standard BASISFIELD::splinefield constructor
      CBFSplineField(const std::vector<unsigned int>& psz, const std::vector<double>& pvxs,
          const std::vector<unsigned int>& pksp, int porder=3);

      unsigned int CoefSz_x() const;
      unsigned int CoefSz_y() const;
      unsigned int CoefSz_z() const;
      /// From base class BASISFIELD::basisfield
      void AsVolume(NEWIMAGE::volume<float>& vol, BASISFIELD::FieldIndex fi=BASISFIELD::FIELD);
      /// From base class BASISFIELD::basisfield
      void SetCoef(const NEWMAT::ColumnVector& pcoef);
      /// Calculate Bending Energy for regularisation purposes
      double BendEnergy() const;
      /// Calculate gradient of the Bending Energy
      NEWMAT::ReturnMatrix BendEnergyGrad() const;
      /// Calculate hessian of the Bending Energy
      std::shared_ptr<MISCMATHS::BFMatrix> BendEnergyHess(
          MISCMATHS::BFMatrixPrecisionType prec) const;
      /// Jte V1
      NEWMAT::ReturnMatrix Jte(const NEWIMAGE::volume<float>&  ima1,
                               const NEWIMAGE::volume<float>&  ima2,
                               const NEWIMAGE::volume<char>    *mask)
                               const;
      /// Jte V2
      NEWMAT::ReturnMatrix Jte(const std::vector<unsigned int>&  deriv,
                               const NEWIMAGE::volume<float>&    ima1,
                               const NEWIMAGE::volume<float>&    ima2,
                               const NEWIMAGE::volume<char>      *mask)
                               const;
      /// Jte V3
      NEWMAT::ReturnMatrix Jte(const NEWIMAGE::volume<float>&    ima,
                               const NEWIMAGE::volume<char>      *mask)
                               const;
      /// Jte V4
      NEWMAT::ReturnMatrix Jte(const std::vector<unsigned int>&  deriv,
                               const NEWIMAGE::volume<float>&    ima,
                               const NEWIMAGE::volume<char>      *mask)
                               const;
      /// JtJ V1
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const NEWIMAGE::volume<float>& ima,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;
      /// JtJ V2
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const NEWIMAGE::volume<float>& ima1,
                                               const NEWIMAGE::volume<float>& ima2,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;
      /// JtJ V3
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const std::vector<unsigned int>& deriv,
                                               const NEWIMAGE::volume<float>& ima,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;
      /// JtJ V4
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const std::vector<unsigned int>& deriv,
                                               const NEWIMAGE::volume<float>& ima1,
                                               const NEWIMAGE::volume<float>& ima2,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;
      /// JtJ V5
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const std::vector<unsigned int>& deriv1,
                                               const NEWIMAGE::volume<float>& ima1,
                                               const std::vector<unsigned int>& deriv2,
                                               const NEWIMAGE::volume<float>& ima2,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;
      /// JtJ V6
      std::shared_ptr<MISCMATHS::BFMatrix> JtJ(const NEWIMAGE::volume<float>& ima1,
                                               const BASISFIELD::basisfield& bf2,
                                               const NEWIMAGE::volume<float>& ima2,
                                               const NEWIMAGE::volume<char> *mask,
                                               MISCMATHS::BFMatrixPrecisionType prec)
                                               const;

    private:
      /// Forward declaration
      class Impl;
      /// Pointer to actual implementation object
      std::unique_ptr<Impl> pimpl_;
  }; // CBFSplineField
} // CBF
#endif // CBF_SPLINE_FIELD_CUH
